from django.db import connections


def is_table_exists(table_name: str, using: str = 'default') -> bool:
    with connections[using].cursor() as cursor:
        return table_name in connections[using].introspection.table_names(cursor)


def is_view_exists(view_name: str, using: str = 'default') -> bool:
    with connections[using].cursor() as cursor:
        views = [
            table.name for table in connections[using].introspection.get_table_list(cursor) if table.type == 'v'
        ]
        return view_name in views



from contextlib import ContextDecorator

VIEW_MIGRATION_CONTEXT = {"is_view_migration": False}


class view_migration_context(ContextDecorator):
    def __enter__(self):
        global VIEW_MIGRATION_CONTEXT
        VIEW_MIGRATION_CONTEXT["is_view_migration"] = True
        return self

    def __exit__(self, *exc):
        global VIEW_MIGRATION_CONTEXT
        VIEW_MIGRATION_CONTEXT["is_view_migration"] = False



import os

from django.conf import settings
import environ
from .dynamic_models_fixtures import *  # noqa

env = environ.Env()
env.read_env(os.path.join(os.path.dirname(os.path.dirname(os.path.abspath(__file__))), '.env'))

pytest_plugins = ()


def pytest_configure():
    """Configure a django settings"""
    postgres_config = {
                'ENGINE': 'django.db.backends.postgresql',
                'NAME': env('POSTGRES_DB'),
                'USER': env('POSTGRES_USER'),
                'PASSWORD': env('POSTGRES_PASSWORD'),
                'HOST': env('POSTGRES_HOST'),
                'PORT': env('POSTGRES_PORT'),
            }
    mysql_config = {
                'ENGINE': 'django.db.backends.mysql',
                'NAME': env('MYSQL_DATABASE'),
                'USER': env('MYSQL_USER'),
                'PASSWORD': env('MYSQL_PASSWORD'),
                'HOST': env('MYSQL_HOST'),
                'PORT': env('MYSQL_PORT'),
                'TEST': {
                    'ENGINE': 'django.db.backends.mysql',
                    'NAME': env('MYSQL_DATABASE'),
                    'USER': env('MYSQL_USER'),
                    'PASSWORD': env('MYSQL_PASSWORD'),
                    'HOST': env('MYSQL_HOST'),
                    'PORT': env('MYSQL_PORT'),
                }
            }
    sqlite = {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': 'test-db'
    }

    settings.configure(
        DATABASES={
            "default": postgres_config,
            # engines
            "postgres": postgres_config,
            "mysql": mysql_config,
            "sqlite": sqlite
        },
        INSTALLED_APPS=[
            "django_db_views",
            "tests.test_app"
        ],
    )



from typing import Union, Callable

from django.db import models, connections, DEFAULT_DB_ALIAS
from django.db.models.base import ModelBase

DBViewsRegistry = {}


class DBViewModelBase(ModelBase):
    def __new__(cls, *args, **kwargs):
        new_class = super().__new__(cls, *args, **kwargs)
        assert new_class._meta.managed is False, "For DB View managed must be se to false"
        DBViewsRegistry[new_class._meta.db_table] = new_class
        return new_class


class DBView(models.Model, metaclass=DBViewModelBase):
    """
        Children should define:
            view_definition - define the view, can be callable or attribute (string)
            view definition can be per db engine.
    """
    view_definition: Union[Callable, str, dict]

    class Meta:
        managed = False
        abstract = True


class DBMaterializedView(DBView):
    class Meta:
        managed = False
        abstract = True

    @classmethod
    def refresh(cls, using=None, concurrently=False):
        """
        concurrently option requires an index and postgres db
        """
        using = using or DEFAULT_DB_ALIAS
        with connections[using].cursor() as cursor:
            if concurrently:
                cursor.execute("REFRESH MATERIALIZED VIEW CONCURRENTLY %s;" % cls._meta.db_table)
            else:
                cursor.execute("REFRESH MATERIALIZED VIEW %s;" % cls._meta.db_table)



import pytest
from django.core.management import call_command

from tests.asserts_utils import is_view_exists


@pytest.fixture(autouse=True)
def backward_compatibility_test_app_settings(settings):
    settings.INSTALLED_APPS += ['tests.backward_compatibility_test_app']
    yield settings
    settings.INSTALLED_APPS.remove('tests.backward_compatibility_test_app')


@pytest.mark.django_db()
@pytest.mark.tag("0.0.9 to 0.0.10")
def test_engine_support_backward_compatibility_migration():
    """Ensures that the initial migration works."""
    assert not is_view_exists("view_for_backward_compatibility_check")
    call_command(
        "migrate", app_label="backward_compatibility_test_app",
        migration_name="0-0-9_to_0-0-10_added_engine_support"
    )
    assert is_view_exists("view_for_backward_compatibility_check")
    call_command(
        "migrate", app_label="backward_compatibility_test_app",
        migration_name="zero"
    )
    assert not is_view_exists("view_for_backward_compatibility_check")



import pytest
from django.core.management import call_command

from tests.asserts_utils import is_view_exists
from tests.decorators import roll_back_schema
from tests.fixturies import temp_migrations_dir, dynamic_models_cleanup  # noqa


@pytest.mark.django_db(transaction=True)
@roll_back_schema
def test_materialized_db_view_based_on_raw_sql_without_dependencies(
        temp_migrations_dir, SimpleMaterializedViewWithoutDependencies
):
    call_command("makeviewmigrations", "test_app")
    assert (temp_migrations_dir / "0001_initial.py").exists()
    call_command("migrate", "test_app")
    assert is_view_exists(SimpleMaterializedViewWithoutDependencies._meta.db_table)
    assert SimpleMaterializedViewWithoutDependencies.objects.all().count() == 1
    current_date_time_from_view_call_1 = SimpleMaterializedViewWithoutDependencies.objects.get().current_date_time
    current_date_time_from_view_call_2 = SimpleMaterializedViewWithoutDependencies.objects.get().current_date_time
    assert current_date_time_from_view_call_1 == current_date_time_from_view_call_2
    # regular refresh
    SimpleMaterializedViewWithoutDependencies.refresh()
    current_date_time_from_view_call_3 = SimpleMaterializedViewWithoutDependencies.objects.get().current_date_time
    assert current_date_time_from_view_call_1 != current_date_time_from_view_call_3
    # backward migration
    call_command("migrate", "test_app", "zero")
    assert not is_view_exists(SimpleMaterializedViewWithoutDependencies._meta.db_table)


@pytest.mark.skip(reason="Future feature.")
@pytest.mark.django_db(transaction=True)
@roll_back_schema
def test_materialized_db_view_based_on_raw_sql_with_indexes(
        temp_migrations_dir, SimpleMaterializedViewWithIndex
):
    call_command("makeviewmigrations", "test_app")
    assert (temp_migrations_dir / "0001_initial.py").exists()
    call_command("migrate", "test_app")
    assert is_view_exists(SimpleMaterializedViewWithIndex._meta.db_table)
    assert SimpleMaterializedViewWithIndex.objects.all().count() == 1
    current_date_time_from_view_call_1 = SimpleMaterializedViewWithIndex.objects.get().current_date_time
    current_date_time_from_view_call_2 = SimpleMaterializedViewWithIndex.objects.get().current_date_time
    assert current_date_time_from_view_call_1 == current_date_time_from_view_call_2
    # regular refresh
    SimpleMaterializedViewWithIndex.refresh()
    current_date_time_from_view_call_3 = SimpleMaterializedViewWithIndex.objects.get().current_date_time
    assert current_date_time_from_view_call_1 != current_date_time_from_view_call_3
    # regular refresh concurrently
    SimpleMaterializedViewWithIndex.refresh(concurrently=True)
    current_date_time_from_view_call_4 = SimpleMaterializedViewWithIndex.objects.get().current_date_time
    assert current_date_time_from_view_call_3 != current_date_time_from_view_call_4
    # backward migration
    call_command("migrate", "test_app", "zero")
    assert not is_view_exists(SimpleMaterializedViewWithIndex._meta.db_table)



from django.apps import apps
from django.db import models
from django.db.models import F, Index
from django.utils import timezone


class QuestionTemplate:
    text = models.CharField(max_length=200)
    created = models.DateTimeField(default=timezone.now)


class ChoiceTemplate:
    question = models.ForeignKey("Question", on_delete=models.CASCADE, related_name='choices')
    text = models.CharField(max_length=200)
    votes = models.IntegerField(default=0)


# DB Views models (inherit from DBView)
class SimpleViewWithoutDependenciesTemplate:
    identifier = models.IntegerField(primary_key=True)
    name = models.TextField()

    view_definition = """
              Select *
                 From  (values (1, 'dummy_1')
                              ,(2, 'dummy_2')
                       ) A(id, name)
            """


class SecondSimpleViewWithoutDependenciesTemplate:
    identifier = models.IntegerField(primary_key=True)
    name = models.TextField()

    view_definition = """
              Select *
                 From  (values (1, 'dummy_1')
                              ,(2, 'dummy_2')
                       ) A(id, name)
            """


class RawViewQuestionStatTemplate:
    question = models.ForeignKey(
        "Question", on_delete=models.DO_NOTHING, related_name="question", primary_key=True
    )
    total_choices = models.IntegerField()

    view_definition = """
      SELECT
          row_number() over () as id,
          q.id as question_id,
          count(*) as total_choices
      FROM test_app_question q
        JOIN test_app_choice c on c.question_id = q.id
      GROUP BY q.id
    """

    class Meta:
        managed = False
        db_table = "question_stat"


class QueryViewQuestionStatTemplate:
    question = models.ForeignKey(
        "Question", on_delete=models.DO_NOTHING, related_name="question", primary_key=True
    )
    total_choices = models.IntegerField()

    @staticmethod
    def view_definition():
        return str(
            apps.get_model(
                'test_app', 'Question'
            ).objects.values("id").annotate(total_choices=models.Count("choices"), question_id=F("id")).query
        )

    class Meta:
        managed = False
        db_table = "question_stat"


class MultipleDBRawViewTemplate:
    identifier = models.IntegerField(primary_key=True)
    name = models.TextField()

    view_definition = {
        "django.db.backends.sqlite3": """
            Select 1 as id, "dummy_1" as name
            UNION
            Select 2 AS id, "dummy_2" AS name
        """,
        "django.db.backends.postgresql": """
              Select *
                 From  (values (1, 'dummy_1')
                              ,(2, 'dummy_2')
                       ) A(id, name)
        """,
    }

    class Meta:
        managed = False
        db_table = "question_stat"


class MultipleDBQueryViewQuestionStatTemplate:
    identifier = models.IntegerField(primary_key=True)
    name = models.TextField()

    @staticmethod
    def view_definition():
        queryset = apps.get_model(
                    'test_app', 'Question'
                ).objects.values("id").annotate(total_choices=models.Count("choices"), question_id=F("id"))
        return {
            "django.db.backends.mysql": str(
                queryset.query.get_compiler("mysql").as_sql()[0]
            ),
            "django.db.backends.postgresql": str(
                queryset.query
            )
        }

    class Meta:
        managed = False
        db_table = "question_stat"


# Materialized View
class SimpleMaterializedViewWithoutDependenciesTemplate:
    current_date_time = models.DateTimeField(primary_key=True)

    view_definition = """
              Select *
                 From  (values (NOW())) A(current_date_time)
            """

    class Meta:
        managed = False
        db_table = "simple_materialized_view_without_dependencies"


class SimpleMaterializedViewWithIndexTemplate:
    current_date_time = models.DateTimeField(primary_key=True)

    view_definition = """
              Select *
                 From  (values (NOW())) A(current_date_time)
            """

    class Meta:
        managed = False
        db_table = "simple_materialized_view_without_dependencies"
        # only django 3.2 +
        indexes = [
            Index(fields=['current_date_time'])
        ]



import functools
from django.core.management import call_command


def roll_back_schema(test_function):
    @functools.wraps(test_function)
    def decorated_test(*args, **kwargs):
        try:
            test_function(*args, **kwargs)
        finally:
            call_command("migrate", "test_app", "zero", database=kwargs.get("database", "default"))
    return decorated_test



from django.utils.deconstruct import deconstructible


class ViewMigration(object):
    DROP_COMMAND_TEMPLATE: str
    CREATE_COMMAND_TEMPLATE: str

    def __init__(self, view_definition: str, table_name: str, engine=None):
        # if provided engine is None, then we are assuming that engine is same as db engine.
        # We do that to keep backward compatibility.
        self.view_engine = engine
        self.view_definition = view_definition
        self.table_name = table_name


class ForwardViewMigrationBase(ViewMigration):
    def __call__(self, apps, schema_editor):
        if self.view_definition:
            if self.view_engine is None or self.view_engine == schema_editor.connection.settings_dict['ENGINE']:
                schema_editor.execute(self.DROP_COMMAND_TEMPLATE % self.table_name)
                schema_editor.execute(self.CREATE_COMMAND_TEMPLATE % (self.table_name, self.view_definition))


class BackwardViewMigrationBase(ViewMigration):
    def __call__(self, apps, schema_editor):
        if self.view_engine is None or self.view_engine == schema_editor.connection.settings_dict['ENGINE']:
            schema_editor.execute(self.DROP_COMMAND_TEMPLATE % self.table_name)
        if self.view_definition:
            if self.view_engine is None or self.view_engine == schema_editor.connection.settings_dict['ENGINE']:
                schema_editor.execute(self.CREATE_COMMAND_TEMPLATE % (self.table_name, self.view_definition))


@deconstructible
class ForwardViewMigration(ForwardViewMigrationBase):
    DROP_COMMAND_TEMPLATE = "DROP VIEW IF EXISTS %s;"
    CREATE_COMMAND_TEMPLATE = "CREATE VIEW %s as %s;"


@deconstructible
class BackwardViewMigration(BackwardViewMigrationBase):
    DROP_COMMAND_TEMPLATE = "DROP VIEW IF EXISTS %s;"
    CREATE_COMMAND_TEMPLATE = "CREATE VIEW %s as %s;"


@deconstructible
class ForwardMaterializedViewMigration(ForwardViewMigrationBase):
    DROP_COMMAND_TEMPLATE = "DROP MATERIALIZED VIEW IF EXISTS %s;"
    CREATE_COMMAND_TEMPLATE = "CREATE MATERIALIZED VIEW %s as %s;"


@deconstructible
class BackwardMaterializedViewMigration(BackwardViewMigrationBase):
    DROP_COMMAND_TEMPLATE = "DROP MATERIALIZED VIEW IF EXISTS %s;"
    CREATE_COMMAND_TEMPLATE = "CREATE MATERIALIZED VIEW %s as %s;"


class DropViewMigration(object):
    DROP_COMMAND_TEMPLATE: str

    def __init__(self, table_name: str, engine=None):
        self.table_name = table_name
        self.view_engine = engine

    def __call__(self, apps, schema_editor):
        if self.view_engine is None or self.view_engine == schema_editor.connection.settings_dict['ENGINE']:
            schema_editor.execute(self.DROP_COMMAND_TEMPLATE % self.table_name)


@deconstructible
class DropMaterializedView(DropViewMigration):
    DROP_COMMAND_TEMPLATE = "DROP MATERIALIZED VIEW IF EXISTS %s;"


@deconstructible
class DropView(DropViewMigration):
    DROP_COMMAND_TEMPLATE = "DROP VIEW IF EXISTS %s;"



from django.db.migrations import operations
from django.db.migrations.state import ModelState

from django_db_views.context_manager import VIEW_MIGRATION_CONTEXT
from django_db_views.db_view import DBView, DBMaterializedView
from django_db_views.migration_functions import ForwardMaterializedViewMigration, ForwardViewMigration


def get_table_engine_name_hash(table_name, engine):
    return f"{table_name}_{engine}".lower()


class DBViewModelState(ModelState):
    def __init__(
            self, *args,
            # Not required cus migrate also load state using clone method that do not provide required by us fields.
            view_engine: str = None,
            view_definition: str = None,
            table_name: str = None,
            base_class=None,
            **kwargs
    ):
        super().__init__(*args, **kwargs)
        if VIEW_MIGRATION_CONTEXT["is_view_migration"]:
            self.view_engine = view_engine
            self.view_definition = view_definition
            self.base_class = base_class
            self.table_name = table_name


class ViewRunPython(operations.RunPython):

    def state_forwards(self, app_label, state):
        if VIEW_MIGRATION_CONTEXT["is_view_migration"]:
            if isinstance(self.code, ForwardMaterializedViewMigration):
                model = DBMaterializedView
            elif isinstance(self.code, ForwardViewMigration):
                model = DBView
            else:
                raise NotImplementedError
            state.add_model(
                DBViewModelState(
                    app_label,
                    # Hash table_name_engine_name to add state model per migration, which are added per engine.
                    get_table_engine_name_hash(self.code.table_name, self.code.view_engine),
                    list(),
                    dict(),
                    # we do not use django bases (they initialize model using that, and broke ViewRegistry),
                    # instead of that we pass bass class in separate argument.
                    tuple(),
                    list(),
                    view_engine=self.code.view_engine,
                    view_definition=self.code.view_definition,
                    base_class=model,
                    table_name=self.code.table_name
                )
            )


class ViewDropRunPython(operations.RunPython):

    def state_forwards(self, app_label, state):
        if VIEW_MIGRATION_CONTEXT["is_view_migration"]:
            state.remove_model(app_label, get_table_engine_name_hash(self.code.table_name, self.code.view_engine))



import pytest
from django.apps import apps
from django.core.management import call_command

from django_db_views.db_view import DBViewsRegistry
from tests.asserts_utils import is_view_exists
from tests.decorators import roll_back_schema
from tests.fixturies import temp_migrations_dir, dynamic_models_cleanup  # noqa


@pytest.mark.django_db(transaction=True)
@roll_back_schema
def test_make_view_migration_for_db_view_based_on_raw_sql_without_dependencies(
        temp_migrations_dir, SimpleViewWithoutDependencies
):
    assert not (temp_migrations_dir / "0001_initial.py").exists()
    assert not is_view_exists(SimpleViewWithoutDependencies._meta.db_table)
    call_command("makeviewmigrations", "test_app")
    assert (temp_migrations_dir / "0001_initial.py").exists()
    call_command("migrate", "test_app")
    assert is_view_exists(SimpleViewWithoutDependencies._meta.db_table)
    assert SimpleViewWithoutDependencies.objects.all().count() == 2
    call_command("migrate", "test_app", "zero")
    assert not is_view_exists(SimpleViewWithoutDependencies._meta.db_table)


@pytest.mark.django_db(transaction=True)
@roll_back_schema
def test_make_view_migration_for_db_view_based_on_raw_sql_with_dependencies_on_other_models(
        temp_migrations_dir, Question, Choice, RawViewQuestionStat
):
    assert not (temp_migrations_dir / "0001_initial.py").exists()
    assert not is_view_exists(RawViewQuestionStat._meta.db_table)
    call_command("makemigrations", "test_app")  # make migrations for regular tables
    call_command("makeviewmigrations", "test_app", name="view_migration")  # make migrations for db view
    assert len(temp_migrations_dir.listdir()) == 3, temp_migrations_dir.listdir()
    assert (temp_migrations_dir / "0001_initial.py").exists()
    assert (temp_migrations_dir / "0002_view_migration.py").exists()
    assert "('test_app', '0001_initial')" in (temp_migrations_dir / "0002_view_migration.py").read()
    call_command("migrate", "test_app")
    assert is_view_exists(RawViewQuestionStat._meta.db_table)
    assert RawViewQuestionStat.objects.all().count() == 0
    # define some instances
    question_1 = Question.objects.create(text="question_1")
    question_2 = Question.objects.create(text="question_2")
    Choice.objects.create(question=question_1, text="choice_1", votes=10)
    Choice.objects.create(question=question_1, text="choice_2", votes=10)
    Choice.objects.create(question=question_2, text="choice_1", votes=10)
    # check whether the model works correctly
    assert RawViewQuestionStat.objects.all().count() == 2
    assert RawViewQuestionStat.objects.all().order_by("question").first().question.text == question_1.text


@pytest.mark.django_db(transaction=True)
@roll_back_schema
def test_make_view_migration_for_db_view_based_on_queryset_with_dependencies_on_other_models(
        temp_migrations_dir, Question, Choice, QueryViewQuestionStat
):
    assert not (temp_migrations_dir / "0001_initial.py").exists()
    assert not is_view_exists(QueryViewQuestionStat._meta.db_table)
    call_command("makemigrations", "test_app")  # make migrations for regular tables
    call_command("makeviewmigrations", "test_app", name="view_migration")  # make migrations for db view
    assert len(temp_migrations_dir.listdir()) == 3
    assert (temp_migrations_dir / "0001_initial.py").exists()
    assert (temp_migrations_dir / "0002_view_migration.py").exists()
    assert "('test_app', '0001_initial')" in (temp_migrations_dir / "0002_view_migration.py").read()
    call_command("migrate", "test_app")
    assert is_view_exists(QueryViewQuestionStat._meta.db_table)
    assert QueryViewQuestionStat.objects.all().count() == 0
    # define some instances
    question_1 = Question.objects.create(text="question_1")
    question_2 = Question.objects.create(text="question_2")
    Choice.objects.create(question=question_1, text="choice_1", votes=10)
    Choice.objects.create(question=question_1, text="choice_2", votes=10)
    Choice.objects.create(question=question_2, text="choice_1", votes=10)
    # check whether the model works correctly
    assert QueryViewQuestionStat.objects.all().count() == 2
    assert QueryViewQuestionStat.objects.all().order_by("question").first().question.text == question_1.text


@pytest.mark.django_db(databases=['postgres', 'sqlite'], transaction=True)
@pytest.mark.parametrize("database", ['postgres', 'sqlite'])
@roll_back_schema
def test_make_view_migration_for_db_raw_view_with_multiple_databases_support(
        temp_migrations_dir, database, MultipleDBRawView
):
    assert not (temp_migrations_dir / "0001_initial.py").exists()
    assert not is_view_exists(MultipleDBRawView._meta.db_table, using=database)
    call_command("makeviewmigrations", "test_app")
    assert (temp_migrations_dir / "0001_initial.py").exists()
    call_command("migrate", "test_app", database=database)
    assert is_view_exists(MultipleDBRawView._meta.db_table, using=database)
    assert MultipleDBRawView.objects.using(database).all().count() == 2


@pytest.mark.django_db(databases=['default', 'postgres', 'mysql'], transaction=True)
@pytest.mark.parametrize("database", ['postgres', 'mysql'])
@roll_back_schema
def test_make_view_migration_for_db_queryset_view_with_multiple_databases_support(
        temp_migrations_dir, database, Question, Choice, MultipleDBQueryViewQuestionStat
):
    assert not (temp_migrations_dir / "0001_initial.py").exists()
    assert not is_view_exists(MultipleDBQueryViewQuestionStat._meta.db_table, using=database)
    call_command("makemigrations", "test_app")
    call_command("makeviewmigrations", "test_app")
    call_command("makemigrations", "test_app")
    assert len(temp_migrations_dir.listdir()) == 3, temp_migrations_dir.listdir()
    assert (temp_migrations_dir / "0001_initial.py").exists()
    call_command("migrate", "test_app", database=database)
    assert is_view_exists(MultipleDBQueryViewQuestionStat._meta.db_table, using=database)
    assert MultipleDBQueryViewQuestionStat.objects.using(database).all().count() == 0
    call_command("migrate", "test_app", "zero", database=database)
    assert not is_view_exists(MultipleDBQueryViewQuestionStat._meta.db_table, using=database)


@pytest.mark.django_db(transaction=True)
@roll_back_schema
def test_move_up_and_down_through_simple_view_stages(
        temp_migrations_dir, SimpleViewWithoutDependencies
):
    call_command("makeviewmigrations", "test_app")
    assert (temp_migrations_dir / "0001_initial.py").exists()
    # made changes
    apps.all_models['test_app']["simpleviewwithoutdependencies"].view_definition = """
              Select *
                 From  (values (3, 'dummy_3')) A(id, name)
            """
    call_command("makeviewmigrations", "test_app", name="second_view_migration")
    assert (temp_migrations_dir / "0002_second_view_migration.py").exists()
    # no changes
    call_command("makeviewmigrations", "test_app", name="third_view_migration")
    assert not (temp_migrations_dir / "0003_third_view_migration.py").exists()
    assert len(temp_migrations_dir.listdir()) == 3
    # we are on stage 2
    call_command("migrate", "test_app")
    assert is_view_exists(SimpleViewWithoutDependencies._meta.db_table)
    assert SimpleViewWithoutDependencies.objects.all().count() == 1
    # go back to stage 1
    call_command("migrate", "test_app", "0001")
    assert is_view_exists(SimpleViewWithoutDependencies._meta.db_table)
    assert SimpleViewWithoutDependencies.objects.all().count() == 2
    # move to stag 1
    call_command("migrate", "test_app", "zero")
    assert not is_view_exists(SimpleViewWithoutDependencies._meta.db_table)


@pytest.mark.django_db(transaction=True)
@roll_back_schema
def test_drop_view(
        temp_migrations_dir, SimpleViewWithoutDependencies
):
    call_command("makeviewmigrations", "test_app")
    assert (temp_migrations_dir / "0001_initial.py").exists()
    del apps.all_models['test_app'][SimpleViewWithoutDependencies.__name__.lower()]
    apps.clear_cache()
    DBViewsRegistry.pop(SimpleViewWithoutDependencies._meta.db_table)
    call_command("makeviewmigrations", "test_app", name="delete_view")
    assert (temp_migrations_dir / "0002_delete_view.py").exists()
    migrations = (temp_migrations_dir / "0002_delete_view.py").read()
    assert "DropView" in migrations
    assert "ViewDropRunPython" in migrations
    # move forward and backward
    call_command("migrate", "test_app", "0001")
    assert is_view_exists(SimpleViewWithoutDependencies._meta.db_table)
    call_command("migrate", "test_app")
    assert not is_view_exists(SimpleViewWithoutDependencies._meta.db_table)
    call_command("migrate", "test_app", "0001")
    assert is_view_exists(SimpleViewWithoutDependencies._meta.db_table)


@pytest.mark.django_db(transaction=True)
@roll_back_schema
def test_drop_view_and_update_existing_view_in_same_migration(
        temp_migrations_dir, SimpleViewWithoutDependencies, SecondSimpleViewWithoutDependencies
):
    call_command("makeviewmigrations", "test_app")
    assert (temp_migrations_dir / "0001_initial.py").exists()
    # Delete existing model
    del apps.all_models['test_app'][SecondSimpleViewWithoutDependencies.__name__.lower()]
    apps.clear_cache()
    DBViewsRegistry.pop(SecondSimpleViewWithoutDependencies._meta.db_table)
    # Modify existing model
    apps.all_models['test_app']["simpleviewwithoutdependencies"].view_definition = """
        Select *
         From  (values (3, 'dummy_3')) A(id, name)
    """
    # Run migration
    call_command("makeviewmigrations", "test_app", name="delete_view_and_update_view")
    assert (temp_migrations_dir / "0002_delete_view_and_update_view.py").exists()
    migrations = (temp_migrations_dir / "0002_delete_view_and_update_view.py").read()
    assert "DropView" in migrations
    assert "ViewDropRunPython" in migrations
    assert "ViewRunPython" in migrations


@pytest.mark.django_db(transaction=True)
@roll_back_schema
def test_drop_view_and_update_existing_view_in_next_migration(
        temp_migrations_dir, SimpleViewWithoutDependencies, SecondSimpleViewWithoutDependencies
):
    call_command("makeviewmigrations", "test_app")
    assert (temp_migrations_dir / "0001_initial.py").exists()
    # Delete existing model
    del apps.all_models['test_app'][SecondSimpleViewWithoutDependencies.__name__.lower()]
    apps.clear_cache()
    DBViewsRegistry.pop(SecondSimpleViewWithoutDependencies._meta.db_table)
    # Run migration
    call_command("makeviewmigrations", "test_app", name="delete_view")
    assert (temp_migrations_dir / "0002_delete_view.py").exists()
    migrations = (temp_migrations_dir / "0002_delete_view.py").read()
    assert "DropView" in migrations
    assert "ViewDropRunPython" in migrations
    assert "ViewRunPython" not in migrations
    # Generate model representaions
    call_command("makemigrations", "test_app", name="create_models")
    # Modify existing model
    apps.all_models['test_app']["simpleviewwithoutdependencies"].view_definition = """
      Select *
         From  (values (3, 'dummy_3')) A(id, name)
    """
    call_command("makeviewmigrations", "test_app", name="update_view")
    assert (temp_migrations_dir / "0004_update_view.py").exists()
    migrations = (temp_migrations_dir / "0004_update_view.py").read()
    assert "ViewRunPython" in migrations
    assert "DropView" not in migrations


@pytest.mark.django_db(databases=['sqlite', 'postgres'], transaction=True)
@pytest.mark.parametrize("database", ['postgres', 'sqlite'])
@roll_back_schema
def test_drop_view_multiple_engines(
        temp_migrations_dir, database, MultipleDBRawView
):
    call_command("makeviewmigrations", "test_app")
    assert (temp_migrations_dir / "0001_initial.py").exists()
    del apps.all_models['test_app'][MultipleDBRawView.__name__.lower()]
    apps.clear_cache()
    DBViewsRegistry.pop(MultipleDBRawView._meta.db_table)
    call_command("makeviewmigrations", "test_app", name="delete_view")
    assert (temp_migrations_dir / "0002_delete_view.py").exists()
    migrations = (temp_migrations_dir / "0002_delete_view.py").read()
    # both migrations are inside.
    assert "engine='django.db.backends.sqlite3" in migrations
    assert "engine='django.db.backends.postgresql" in migrations
    call_command("migrate", "test_app", "0001", database=database)
    assert is_view_exists(MultipleDBRawView._meta.db_table, using=database)
    call_command("migrate", "test_app", database=database)
    assert not is_view_exists(MultipleDBRawView._meta.db_table, using=database)
    call_command("migrate", "test_app", "0001", database=database)
    assert is_view_exists(MultipleDBRawView._meta.db_table, using=database)



import pytest
from django.db import models

from tests.test_app.models import QuestionTemplate, SimpleViewWithoutDependenciesTemplate, ChoiceTemplate, \
    RawViewQuestionStatTemplate, QueryViewQuestionStatTemplate, MultipleDBRawViewTemplate, \
    MultipleDBQueryViewQuestionStatTemplate, SimpleMaterializedViewWithoutDependenciesTemplate, \
    SimpleMaterializedViewWithIndexTemplate, SecondSimpleViewWithoutDependenciesTemplate


def define_model(template_class, parent):
    attributes = get_declared_class_attributes(template_class)
    attrs = {
        **attributes,
        '__module__': 'tests.test_app.models'
    }
    return type(template_class.__name__.replace("Template", ""), (parent,), attrs)


def get_declared_class_attributes(cls) -> dict:
  return {key: value for key, value in cls.__dict__.items() if not key.startswith('__')}


@pytest.fixture
def Question():
    return define_model(QuestionTemplate, models.Model)


@pytest.fixture()
def Choice():
    return define_model(ChoiceTemplate, models.Model)


@pytest.fixture
def SimpleViewWithoutDependencies():
    from django_db_views.db_view import DBView
    return define_model(SimpleViewWithoutDependenciesTemplate, DBView)


@pytest.fixture
def SecondSimpleViewWithoutDependencies():
    from django_db_views.db_view import DBView
    return define_model(SecondSimpleViewWithoutDependenciesTemplate, DBView)


@pytest.fixture
def RawViewQuestionStat():
    from django_db_views.db_view import DBView
    return define_model(RawViewQuestionStatTemplate, DBView)


@pytest.fixture
def QueryViewQuestionStat():
    from django_db_views.db_view import DBView
    return define_model(QueryViewQuestionStatTemplate, DBView)


@pytest.fixture
def MultipleDBRawView():
    from django_db_views.db_view import DBView
    return define_model(MultipleDBRawViewTemplate, DBView)


@pytest.fixture
def MultipleDBQueryViewQuestionStat():
    from django_db_views.db_view import DBView
    return define_model(MultipleDBQueryViewQuestionStatTemplate, DBView)


@pytest.fixture
def SimpleMaterializedViewWithoutDependencies():
    from django_db_views.db_view import DBMaterializedView
    return define_model(SimpleMaterializedViewWithoutDependenciesTemplate, DBMaterializedView)


@pytest.fixture
def SimpleMaterializedViewWithIndex():
    from django_db_views.db_view import DBMaterializedView
    return define_model(SimpleMaterializedViewWithIndexTemplate, DBMaterializedView)



import importlib

import pytest
from django.apps import apps


@pytest.fixture(autouse=True, scope='function')
def temp_migrations_dir(settings, tmpdir, mocker):
    migrations_dir = tmpdir.mkdir("migrations")
    init_file = (migrations_dir / "__init__.py")
    init_file.write_text("", encoding="utf-8")
    mocker.patch("django.db.migrations.writer.MigrationWriter.basedir", migrations_dir.strpath)

    def new_module_import(module_name):
        if module_name == 'tests.test_app.migrations':
            spec = importlib.util.spec_from_file_location(module_name, init_file)
            return importlib.util.module_from_spec(spec)
        elif 'tests.test_app.migrations' in module_name:
            spec = importlib.util.spec_from_file_location(module_name, migrations_dir / f"{module_name.split('.')[-1]}.py")
            module = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(module)
            return module
        return importlib.import_module(module_name)
    mocker.patch(
        "django.db.migrations.loader.import_module",
        new_module_import
    )
    return migrations_dir


@pytest.fixture(autouse=True, scope='function')
def dynamic_models_cleanup():
    try:
        yield None
    finally:
        # We delete all dynamically created models
        test_app_models = list(apps.all_models['test_app'].keys())
        for model_name in test_app_models:
            del apps.all_models['test_app'][model_name]
        apps.clear_cache()
